from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, TypedDict

from typing_extensions import Unpack

from .exceptions import EncodeError

if TYPE_CHECKING:
    from pydantic import BaseModel
    from pydantic.main import IncEx

    from .types import Encoder


class Options(TypedDict, total=False):
    indent: int | None
    include: IncEx
    exclude: IncEx
    context: Any | None
    by_alias: bool
    exclude_unset: bool
    exclude_defaults: bool
    exclude_none: bool
    round_trip: bool
    warnings: bool | Literal["none", "warn", "error"]
    serialize_as_any: bool


class JsonEncoder:
    CONTENT_TYPE: str = "application/json"

    def __init__(self, **options: Unpack[Options]) -> None:
        self.options = options

    def encode(self, data: BaseModel) -> bytes:
        try:
            return data.model_dump_json(**self.options).encode("utf-8")
        except Exception as e:
            raise EncodeError from e


DEFAULT_ENCODER: Encoder = JsonEncoder(by_alias=True)



from __future__ import annotations

import asyncio
from abc import ABC
from typing import TYPE_CHECKING, Annotated, Any, Callable

import anyio
from nats.aio.client import Client
from nats.aio.msg import Msg as NatsMsg
from nats.js import JetStreamContext, api
from nats.js.api import ConsumerConfig
from nats.js.errors import FetchTimeoutError, KeyNotFoundError
from pydantic import AnyUrl, Field, UrlConstraints

from eventiq.broker import R, UrlBroker
from eventiq.exceptions import BrokerError
from eventiq.results import ResultBackend
from eventiq.settings import UrlBrokerSettings
from eventiq.utils import to_float

if TYPE_CHECKING:
    from collections.abc import Awaitable

    from eventiq.types import ID, DecodedMessage

NatsUrl = Annotated[AnyUrl, UrlConstraints(allowed_schemes=["nats"])]


class NatsSettings(UrlBrokerSettings[NatsUrl]):
    auto_flush: bool = True


class JetStreamSettings(NatsSettings):
    jetstream_options: dict[str, Any] = Field({})
    kv_options: dict[str, Any] = Field({})


if TYPE_CHECKING:
    from anyio.streams.memory import MemoryObjectSendStream
    from nats.js.kv import KeyValue

    from eventiq import Consumer


class AbstractNatsBroker(UrlBroker[NatsMsg, R], ABC):
    """:param auto_flush: auto flush messages on publish
    :param auto_flush: auto flush on publish
    :param kwargs: options for base class
    """

    protocol = "nats"
    WILDCARD_ONE = "*"
    WILDCARD_MANY = ">"

    def __init__(
        self,
        *,
        auto_flush: bool = True,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.client = Client()
        self._auto_flush = auto_flush
        self.connection_options.setdefault("max_reconnect_attempts", 10)
        for k in ("error", "closed", "reconnected", "disconnected"):
            self.connection_options.setdefault(f"{k}_cb", self._default_cb(k))

    def _default_cb(
        self, message: str
    ) -> Callable[[Exception | None], Awaitable[None]]:
        async def wrapped(error: Exception | None = None) -> None:
            self.logger.warning(message)
            if error:
                self.logger.error(error)

        return wrapped

    @staticmethod
    def decode_message(raw_message: NatsMsg) -> DecodedMessage:
        return raw_message.data, raw_message.headers

    @staticmethod
    def get_message_metadata(raw_message: NatsMsg) -> dict[str, str]:
        try:
            return {
                "messaging.nats.sequence.consumer": str(
                    raw_message.metadata.sequence.consumer,
                ),
                "messaging.nats.sequence.stream": str(
                    raw_message.metadata.sequence.stream,
                ),
                "messaging.nats.num_delivered": str(raw_message.metadata.num_delivered),
            }
        except Exception:
            return {}

    async def connect(self) -> None:
        if not self.client.is_connected:
            await self.client.connect(self.url, **self.connection_options)

    async def disconnect(self) -> None:
        if self.client.is_connected:
            await self.client.drain()
            await self.client.close()

    async def flush(self) -> None:
        await self.client.flush()

    @property
    def is_connected(self) -> bool:
        return self.client.is_connected

    async def ack(self, raw_message: NatsMsg) -> None:
        await raw_message.ack()

    async def nack(self, raw_message: NatsMsg, delay: int | None = None) -> None:
        await raw_message.nak(delay=delay)


class NatsBroker(AbstractNatsBroker[None]):
    Settings = NatsSettings

    async def sender(
        self,
        group: str,
        consumer: Consumer,
        send_stream: MemoryObjectSendStream,
    ) -> None:
        queue = "" if consumer.dynamic else f"{group}:{consumer.name}"
        subscription = await self.client.subscribe(
            subject=self.format_topic(consumer.topic),
            queue=queue,
        )
        try:
            async with send_stream:
                async for message in subscription.messages:
                    await send_stream.send(message)
        finally:
            with anyio.move_on_after(1, shield=True):
                if consumer.dynamic:
                    await subscription.unsubscribe()
            self.logger.info("Sender finished for %s", consumer.name)

    async def publish(
        self,
        topic: str,
        body: bytes,
        *,
        headers: dict[str, str],
        reply: str = "",
        flush: bool = False,
        **kwargs: Any,
    ) -> None:
        await self.client.publish(topic, body, headers=headers, reply=reply)
        if self._auto_flush or flush:
            await self.flush()


class JetStreamBroker(
    AbstractNatsBroker[api.PubAck],
    ResultBackend[NatsMsg, api.PubAck],
):
    """NatsBroker with JetStream enabled
    :param jetstream_options: additional options passed to nc.jetstream(...)
    :param kv_options: options for nats KV initialization.
    :param kwargs: all other options for base classes NatsBroker, Broker.
    """

    Settings = JetStreamSettings
    kv_error = "KeyVal not initialized"

    def __init__(
        self,
        *,
        jetstream_options: dict[str, Any] | None = None,
        kv_options: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.jetstream_options = jetstream_options or {}
        self.js = JetStreamContext(self.client, **self.jetstream_options)
        self.kv_options = kv_options or {}
        self._kv: KeyValue | None = None

    async def init_storage(self) -> None:
        self._kv = await self.js.create_key_value(**self.kv_options)

    @property
    def kv(self) -> KeyValue:
        if self._kv is None:
            raise BrokerError(self.kv_error)
        return self._kv

    async def get_result(self, key: str) -> bytes | None:
        try:
            data = await self.kv.get(key)
            return data.value  # noqa: TRY300
        except KeyNotFoundError:
            self.logger.warning("Key %s not found", key)

    async def store_result(self, key: str, result: bytes) -> None:
        await self.kv.put(key, result)

    async def publish(
        self,
        topic: str,
        body: bytes,
        *,
        headers: dict[str, str],
        message_id: ID,
        timeout: float | None = None,
        stream: str | None = None,
        **kwargs: Any,
    ) -> api.PubAck:
        if "Nats-Msg-Id" not in headers:
            headers["Nats-Msg-Id"] = str(message_id)
        response = await self.js.publish(
            topic, payload=body, timeout=timeout, stream=stream, headers=headers
        )
        if self._auto_flush:
            await self.flush()
        return response

    async def sender(
        self,
        group: str,
        consumer: Consumer,
        send_stream: MemoryObjectSendStream,
    ) -> None:
        config_kwargs: dict[str, Any] = {
            "name": f"{group}:{consumer.name}",
            "ack_wait": (to_float(consumer.timeout) or self.default_consumer_timeout)
            + 30,
            "max_ack_pending": 10_000,
        }
        for key in ConsumerConfig.__dataclass_fields__:
            if key in consumer.options:
                config_kwargs[key] = consumer.options[key]
        config = ConsumerConfig(**config_kwargs)
        fetch_timeout = consumer.options.get("fetch_timeout", 10)
        heartbeat = consumer.options.get("heartbeat", 0.1)
        durable = None if consumer.dynamic else f"{group}:{consumer.name}"
        subscription = await self.js.pull_subscribe(
            subject=self.format_topic(consumer.topic),
            durable=durable,
            config=config,
        )
        try:
            async with send_stream:
                while True:
                    try:
                        batch = consumer.concurrency - len(
                            send_stream._state.buffer  # noqa: SLF001
                        )
                        if batch == 0:
                            await asyncio.sleep(0.1)
                            continue
                        self.logger.debug("Fetching %d messages", batch)
                        messages = await subscription.fetch(
                            batch=batch,
                            timeout=fetch_timeout,
                            heartbeat=heartbeat,
                        )
                        for message in messages:
                            await send_stream.send(message)
                    except FetchTimeoutError:
                        pass
        finally:
            if consumer.dynamic:
                await subscription.unsubscribe()
            self.logger.info("Stopped sender for consumer: %s", consumer.name)

    def should_nack(self, raw_message: NatsMsg) -> bool:
        return raw_message.metadata.num_delivered <= 3

    def get_num_delivered(self, raw_message: NatsMsg) -> int | None:
        return raw_message.metadata.num_delivered



from __future__ import annotations

from typing import TYPE_CHECKING, Protocol

from eventiq.middleware import CloudEventType, Middleware

if TYPE_CHECKING:
    from eventiq import Consumer, Service


class Limiter(Protocol):
    """
    This is protocol for user to be implemented.
    For single instance rate limits `aiolimiter` (https://github.com/mjpieters/aiolimiter)
    package provides compatible interface.
    For distributed rate limits redis can be used.
    """

    async def acquire(self) -> None: ...


class RateLimitMiddleware(Middleware[CloudEventType]):
    def __init__(self, service: Service, limiter: Limiter | None = None) -> None:
        super().__init__(service)
        self.limiter = limiter

    async def before_process_message(
        self, *, consumer: Consumer, message: CloudEventType
    ) -> None:
        limiter: Limiter = consumer.options.get("limiter", self.limiter)

        if limiter is not None:
            await limiter.acquire()



from typing import Any, Generic, Literal, Optional, TypeVar, Union

from pydantic.networks import AnyUrl
from pydantic_settings import BaseSettings, SettingsConfigDict

from .imports import ImportedType
from .types import Decoder, DefaultAction, Encoder


class BrokerSettings(BaseSettings):
    name: str = "default"
    description: Optional[str] = None
    default_on_exc: DefaultAction = "nack"
    default_consumer_timeout: int = 300
    validate_error_delay: Optional[int] = 3600 * 12

    model_config = SettingsConfigDict(env_prefix="BROKER_")


Url = TypeVar("Url", bound=Union[AnyUrl, str])


class UrlBrokerSettings(BrokerSettings, Generic[Url]):
    url: Url
    connection_options: dict[str, Any] = {}


class ServiceSettings(BaseSettings):
    name: str
    title: Optional[str] = None
    version: str = "0.1.0"
    description: str = ""
    asyncapi: Literal["3.0.0"] = "3.0.0"
    encoder: Optional[ImportedType[Encoder]] = None
    decoder: Optional[ImportedType[Decoder]] = None



from __future__ import annotations

import asyncio
import functools
import re
from collections.abc import Awaitable
from datetime import datetime, timedelta, timezone
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Literal,
    TypeVar,
    cast,
    get_type_hints,
    overload,
)
from urllib.parse import urlparse

from anyio import to_thread
from typing_extensions import ParamSpec, TypeGuard

if TYPE_CHECKING:
    from eventiq.types import Timeout

P = ParamSpec("P")
R = TypeVar("R", bound=Any)


TOPIC_PATTERN = re.compile(r"{\w+}")
TOPIC_SPECIAL_CHARS = {"{", "}", "*", ">"}


def utc_now() -> datetime:
    return datetime.now(tz=timezone.utc)


def to_async(func: Callable[P, R]) -> Callable[P, Awaitable[R]]:
    @functools.wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> Awaitable[R]:
        if not kwargs:
            return to_thread.run_sync(func, *args)
        return to_thread.run_sync(functools.partial(func, *args, **kwargs))

    return wrapper


def get_safe_url(url: str) -> str:
    parsed = urlparse(url)
    if parsed.username and parsed.password:
        parsed = parsed._replace(
            netloc="{}:{}@{}:{}".format(
                parsed.username or "",
                "*****",
                parsed.hostname,
                parsed.port,
            ),
        )
    return parsed.geturl()


def resolve_message_type_hint(func: Callable) -> type[Any] | None:
    try:
        return func.__annotations__["message"]
    except (AttributeError, KeyError):
        pass
    hints = get_type_hints(func)
    if "message" in hints:
        return hints["message"]
    hints.pop("return", None)
    try:
        return next(iter(hints.values()))
    except StopIteration:
        return None


def format_topic(topic: str, wildcard_one: str, wildcard_many: str) -> str:
    result = []

    for k in topic.split("."):
        if re.fullmatch(TOPIC_PATTERN, k):
            result.append(wildcard_one)
        elif k in {"*", ">"}:
            result.append(wildcard_many)
        else:
            result.append(k)
    return ".".join(filter(None, result))


def get_topic_regex(topic: str) -> str:
    result = []

    for k in topic.split("."):
        if re.fullmatch(TOPIC_PATTERN, k):
            result.append(r"\w+")

        elif k in {"*", ">"}:
            result.append(r"*")
        else:
            result.append(k)
    return r"^{}$".format(r"\.".join(result))


@overload
def to_float(timeout: Timeout) -> float: ...


@overload
def to_float(timeout: None) -> None: ...


def to_float(timeout: Timeout | None) -> float | None:
    if timeout is None:
        return None
    if isinstance(timeout, timedelta):
        return timeout.total_seconds()
    return float(timeout)


def get_annotation(value: str) -> type:
    return cast(type, Literal[value])


T = TypeVar("T")


AwaitableCallable = Callable[..., Awaitable[T]]


@overload
def is_async_callable(obj: AwaitableCallable[T]) -> TypeGuard[AwaitableCallable[T]]: ...


@overload
def is_async_callable(obj: Any) -> TypeGuard[AwaitableCallable[Any]]: ...


def is_async_callable(obj: Any) -> Any:
    while isinstance(obj, functools.partial):
        obj = obj.func

    return asyncio.iscoroutinefunction(obj) or (
        callable(obj) and asyncio.iscoroutinefunction(obj.__call__)
    )



from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Any, Generic, Optional, TypeVar
from uuid import UUID, uuid4

from pydantic import (
    AnyUrl,
    BaseModel,
    Field,
    PrivateAttr,
    field_validator,
    model_validator,
)
from pydantic.fields import FieldInfo, _FieldInfoInputs
from typing_extensions import Self

from .types import Parameter
from .utils import TOPIC_SPECIAL_CHARS, get_annotation, get_topic_regex, utc_now

if TYPE_CHECKING:
    from .service import Service

D = TypeVar("D", bound=Any)


class CloudEvent(BaseModel, Generic[D]):
    """Base Schema for all messages."""

    specversion: str = Field("1.0", description="CloudEvents specification version")
    content_type: Optional[str] = Field(
        None,
        alias="datacontenttype",
        description="Message content type",
    )
    id: UUID = Field(default_factory=uuid4, description="Event ID", repr=True)
    time: datetime = Field(default_factory=utc_now, description="Event created time")
    topic: str = Field(
        None,
        alias="subject",
        description="Message subject (topic)",
        validate_default=True,
    )
    type: str = Field(None, description="Event type", validate_default=True)
    source: Optional[str] = Field(None, description="Event source (app)")
    data: D = Field(..., description="Event payload")
    dataschema: Optional[AnyUrl] = Field(None, description="Data schema URI")

    _raw: Optional[Any] = PrivateAttr(None)
    _headers: dict[str, str] = PrivateAttr({})
    _service: Optional["Service"] = PrivateAttr(None)

    def __init_subclass__(
        cls,
        abstract: bool = False,
        topic: Optional[str] = None,
        validate_topic: bool = False,
        **kwargs: Any,
    ) -> None:
        if not abstract and topic:
            kw: _FieldInfoInputs = {
                "alias": "subject",
                "description": "Message subject",
                "validate_default": True,
            }
            if any(k in topic for k in TOPIC_SPECIAL_CHARS):
                kw.update(
                    {
                        "annotation": str,
                        "default": topic,
                    },
                )
                if validate_topic:
                    kw["pattern"] = get_topic_regex(topic)
            else:
                kw.update(
                    {
                        "annotation": get_annotation(topic),
                        "default": topic,
                    },
                )

            cls.model_fields["topic"] = FieldInfo(**kw)
        super().__init_subclass__(**kwargs)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, CloudEvent):
            return False
        return self.id == other.id

    def __hash__(self) -> int:
        return hash(self.id)

    def __str__(self) -> str:
        return (
            f"{type(self).__name__}(type={self.type}, topic={self.topic}, id={self.id})"
        )

    def __repr__(self) -> str:
        return str(self)

    @classmethod
    def get_default_topic(cls) -> Optional[str]:
        return cls.model_fields["topic"].get_default()

    @field_validator("type", mode="before")
    @classmethod
    def get_default_type(cls, value: Any) -> str:
        if value is None:
            return cls.__name__
        return str(value)

    @property
    def raw(self) -> Any:
        if self._raw is None:
            msg = "raw property accessible only for incoming messages"
            raise ValueError(msg)
        return self._raw

    @property
    def service(self) -> "Service":
        if self._service is None:
            msg = "Service not set"
            raise ValueError(msg)
        return self._service

    def set_context(
        self, service: "Service", raw: Any, headers: dict[str, str]
    ) -> None:
        self._service = service
        self._raw = raw
        self._headers = headers

    def model_dump(self, **kwargs: Any) -> dict[str, Any]:
        kwargs.setdefault("by_alias", True)
        kwargs.setdefault("exclude_none", True)
        return super().model_dump(**kwargs)

    @classmethod
    def new(
        cls, obj: D, *, headers: Optional[dict[str, str]] = None, **kwargs: Any
    ) -> Self:
        self = cls(data=obj, **kwargs)
        if headers:
            self._headers.update(headers)
        return self

    @property
    def age(self) -> timedelta:
        return utc_now() - self.time

    @property
    def headers(self) -> dict[str, str]:
        return self._headers

    model_config = {
        "use_enum_values": True,
        "populate_by_name": True,
        "extra": "allow",
        "arbitrary_types_allowed": True,
    }


class Publishes(BaseModel):
    type: type[CloudEvent]
    topic: str = Field(None)
    parameters: dict[str, Parameter] = {}
    tags: list[str] = []
    summary: str = ""
    asyncapi_extra: dict[str, Any] = {}

    @model_validator(mode="after")
    def validate_topic(self) -> Self:
        if self.topic is None:
            topic = self.type.get_default_topic()
            if topic is None:
                msg = "Topic is required"
                raise ValueError(msg)
            self.topic = topic
        return self

    model_config = {
        "populate_by_name": True,
        "extra": "allow",
        "arbitrary_types_allowed": True,
    }


class Event(CloudEvent[D], abstract=True):
    pass


class Command(CloudEvent[D], abstract=True):
    pass


class Query(CloudEvent[D], abstract=True):
    pass



from __future__ import annotations

from contextlib import AbstractAsyncContextManager
from datetime import timedelta
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Literal,
    Optional,
    Protocol,
    TypeVar,
    Union,
    overload,
    runtime_checkable,
)
from uuid import UUID

from pydantic import BaseModel, TypeAdapter
from typing_extensions import Concatenate, ParamSpec, TypedDict

if TYPE_CHECKING:
    from .consumer import GenericConsumer
    from .middleware import Middleware
    from .models import CloudEvent, Publishes
    from .service import Service


Undefinded: Any = object()

ID = Union[UUID, str]

Message = TypeVar("Message", bound=Any)
DefaultAction = Literal["ack", "nack"]
DecodedMessage = tuple[bytes, Optional[dict[str, str]]]

T = TypeVar("T", bound=BaseModel)
Seconds = Union[int, float]
Timeout = Union[Seconds, timedelta]
RawData = Union[str, bytes, bytearray]

CloudEventType = TypeVar("CloudEventType", bound="CloudEvent")
AnyType: TypeAdapter = TypeAdapter(Any)
State = dict[Union[type, str], Any]

PreparedMessage = tuple[str, bytes, dict[str, Any]]
Lifespan = Callable[["Service"], AbstractAsyncContextManager[Optional[State]]]


P = ParamSpec("P")

MessageHandler = Union[
    type["GenericConsumer"], Callable[Concatenate[CloudEventType, P], Any]
]


class MiddlewareType(Protocol[P]):
    def __call__(
        self, service: Service, *args: P.args, **kwargs: P.kwargs
    ) -> Middleware: ...


class Publisher(Protocol):
    async def __call__(
        self,
        message: CloudEvent,
        topic: str | None = None,
        headers: dict[str, str] | None = None,
        encoder: Encoder | None = None,
        **kwargs: Any,
    ) -> Any: ...


@runtime_checkable
class Encoder(Protocol):
    CONTENT_TYPE: str

    def encode(self, data: BaseModel) -> bytes: ...


@runtime_checkable
class Decoder(Protocol[T]):
    CONTENT_TYPE: str

    @overload
    def decode(self, data: RawData, as_type: type[T]) -> T: ...

    @overload
    def decode(self, data: RawData, as_type: None = None) -> Any: ...


class Parameter(TypedDict, total=False):
    enum: list[str]
    default: str
    description: str
    examples: list[str]
    location: str


class RetryStrategy(Protocol):
    def maybe_retry(
        self,
        service: Service,
        message: CloudEvent,
        exc: Exception,
    ) -> None: ...


class ConsumerGroupOptions(TypedDict, total=False):
    topic: str
    timeout: Timeout
    dynamic: bool
    tags: list[str]
    encoder: Encoder
    decoder: Decoder
    description: str
    concurrency: int
    retry_strategy: RetryStrategy
    publishes: list[Publishes]
    parameters: dict[str, Parameter]
    asyncapi_extra: dict[str, Any]



from typing import Any
from unittest.mock import MagicMock

from eventiq import CloudEvent, Service
from eventiq.backends.stub import StubBroker


def test_service(service: Service):
    assert isinstance(service, Service)
    assert isinstance(service.broker, StubBroker)
    assert service.name == "test_service"


async def test_consumer_called(
    running_service: Service, ce: CloudEvent, mock_consumer: Any
):
    await running_service.publish(ce)
    mock_consumer.assert_called_once_with(ce)


async def test_bulk_publish(
    running_service: Service, ce: CloudEvent, mock_consumer: MagicMock
):
    messages = [ce] * 10
    for message in messages:
        assert "Content-Type" not in message.headers
        assert message.source is None
    await running_service.bulk_publish(messages)
    assert mock_consumer.call_count == 10
    for message in messages:
        assert message.headers["Content-Type"] == "application/json"
        assert message.source == running_service.name



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from eventiq.middleware import CloudEventType, Middleware

if TYPE_CHECKING:
    from eventiq import Consumer, Service
    from eventiq.exceptions import Fail


class DeadLetterQueueMiddleware(Middleware[CloudEventType]):
    def __init__(
        self,
        service: Service,
        topic: str = "dlx",
        **kwargs: Any,
    ) -> None:
        super().__init__(service)
        self.topic = topic
        self.kwargs = kwargs

    async def after_fail_message(
        self,
        *,
        consumer: Consumer,
        message: CloudEventType,
        exc: Fail,
    ) -> None:
        dlx_message = message.copy()
        dlx_message.headers.update(
            {
                "exc-reason": exc.reason,
            }
        )
        await self.service.publish(dlx_message, topic=self.topic, **self.kwargs)



RET_VAL = 42


async def test_consumer_process(test_consumer, ce):
    res = await test_consumer.process(ce)
    assert res == RET_VAL


async def test_generic_consumer_process(generic_test_consumer, ce):
    res = await generic_test_consumer.process(ce)
    assert res == RET_VAL



import pytest
from pydantic import ValidationError

from eventiq import CloudEvent


@pytest.fixture()
def test_event_cls():
    class TestEvent(CloudEvent[str], topic="events.{type}", validate_topic=True):
        pass

    return TestEvent


@pytest.fixture()
def test_command_cls():
    class TestCommand(CloudEvent[str], topic="commands.command_a"):
        pass

    return TestCommand


def test_get_default_topic(test_event_cls):
    assert test_event_cls.get_default_topic() == "events.{type}"


@pytest.mark.parametrize(
    "topic",
    ["some_random_string", "events.type.subtype", "events."],
)
def test_event_incorrect_topic(test_event_cls, topic):
    with pytest.raises(ValidationError):
        test_event_cls(topic=topic, data="test_data")


@pytest.mark.parametrize("topic", ["events.type", "events.type2"])
def test_correct_topic(test_event_cls, topic):
    test_event_cls(topic=topic, data="test_data")


@pytest.mark.parametrize(
    "topic",
    ["some_random_string", "events.type.subtype", "events."],
)
def test_command_incorrect_topic(test_command_cls, topic):
    with pytest.raises(ValidationError):
        test_command_cls(topic=topic, data="test_data")


def test_command_correct_topic(test_command_cls):
    test_command_cls(data="test_data")
    test_command_cls(data="test_data", topic="commands.command_a")


def test_untyped_model():
    class UserEvent(CloudEvent): ...

    u1 = UserEvent.new(
        {"name": "John"},
        type="UserCreated",
        topic="events.users.created",
    )
    assert isinstance(u1, UserEvent)
    assert u1.type == "UserCreated"
    u2 = UserEvent.new(
        {"name": "John"},
        type="UserUpdated",
        topic="events.users.updated",
    )
    assert u2.type == "UserUpdated"
    assert isinstance(u2, UserEvent)
    u3 = UserEvent.new(
        {"name": "John"},
        type="UserDeleted",
        topic="events.users.deleted",
    )
    assert isinstance(u3, UserEvent)
    assert u3.type == "UserDeleted"
